import numpy as np
import time
from math import exp, sqrt, pi, cos, sin
import itertools
from mpi4py import MPI

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from plot_serial import draw_once

import sys

import argparse

import itertools
from plot_serial import draw_once
from matplotlib.backends.backend_pdf import PdfPages
from collections import deque as dq

import time

#Helper to draw a normal
def norm_sample(mu,sigma):
	return np.random.normal(loc=mu, scale = sigma)

#Compute the force vector
def v_vector(r1, r2, c1, c2, D, v_const):
	v_abs = v_const * (r1 + r2 - D) / 2.0
	return v_abs * (c1 - c2)
	
#Helper to compute distance
def distance(c1,c2):
	return sqrt(np.dot((c1 - c2), (c1 - c2).T))
	
#For division, choose a displacement a small distance away, and then return two centers
def divide(centroid,displacement):
	theta = np.random.random() * 2 * pi
	direction = np.array([cos(theta),sin(theta)],dtype=np.float32)
	return (centroid - direction * displacement), (centroid + direction * displacement) 

#Helper to determine whether two nodes are within some threshold of each other
def is_neighbor(node1, node2, cell_data, glob_thresh):
	c1r = cell_data[node1][0:3]
	c2r = cell_data[node2][0:3]
	c1 = c1r[:2]
	r1 = c1r[2]
	c2 = c2r[:2]
	r2 = c2r[2]
	D =  distance(c1, c2)
	# glob_thresh = 0.0
	if (r1 + r2 ) + glob_thresh > D:
		return True
	else:
		return False

#Construct a new graph of interactions using BFS to search for new neighbors
def make_graph_bfs2(graph, cell_data, glob_thresh):
	new_graph = {}

	for node in graph:
		new_adj_list = []
		
		q, visited = dq(), [node]
		
		#Append its neighbors to the queue to be visited
		for n in graph[node]:
			q.append((n,1))
			
		#Flag whether a level has been fully visited yet
		level_flags = {0: 1, 1: 0}
		max_dist = 1

		while len(q) > 0:
			#Current node being visited, along with its level
			v, d = q.popleft()

			if d > max_dist:
				if level_flags[d-1] == 0:
					break
				else:
					level_flags[d] = 0
					max_dist = d

			if v not in visited:
				visited.append(v)
				if is_neighbor(node, v, cell_data, glob_thresh):
					new_adj_list.append(v)
					if level_flags[d] == 0:
						level_flags[d] = 1
				#Add neighbors of n into the queue for BFS
				for n in graph[v]:
					if n not in visited:
						q.append((n, d+1))
		new_graph[node] = list(set(new_adj_list))
	return new_graph
	
#Given an initial array for Cell_Data, generate dictionary formats that can interface with BFS
def generate_graph(cell_list):
	new_graph = {}
	new_cell_data = {}
	new_force_vector = {}
	
	num_feats = cell_list.shape[1]
	
	for i in range(cell_list.shape[0]):
		new_cell_data[str(i)] = cell_list[i,0:num_feats-2]
		new_force_vector[str(i)] = cell_list[i,num_feats-2:]
		new_adj_list = [];
		
		#Search through for the adjacency list
		for j in range(cell_list.shape[0]):
			if i != j:
				c1r = cell_data[i,0:3]
				c2r = cell_data[j,0:3]
				c1 = c1r[:2]
				r1 = c1r[2]
				c2 = c2r[:2]
				r2 = c2r[2]
				D =  distance(c1, c2)
				if (r1 + r2) > D:
					new_adj_list.append(str(j))
					
		new_graph[str(i)] = list(set(new_adj_list))
		
	return new_graph, new_cell_data, new_force_vector

#RUNMODE 0: Naive Serial Implementation
def simulate_naive_serial(init_cell_data,args):
	dt = args.dt
	t_elapsed = 0
	t_end = args.time
	
	displacement = args.disp
	v_const = 25.0
	
	grow_const = args.grow
	mu = args.rmax
	sigma = args.rsd
 
	# plotting stuff
	if args.draw:
		plt.ion()
		fig = plt.figure(figsize=(16, 12)).add_subplot(111, aspect='equal')
		plt.xlim(-15,15)
		plt.ylim(-15,15)
	
	cell_data = init_cell_data

	num_cells = cell_data.shape[0];
	
	
	while t_elapsed < t_end:
		#Grow and Divide
		for i in xrange(cell_data.shape[0]):
			(r, r_max, t_start) = cell_data[i, 2:5]
			r = grow_const * exp((t_elapsed - t_start)/2.0)
			
			if r < r_max:
				cell_data[i,2] = r #grow
			else:
				# divide
				num_cells +=1
				(new_pos,child_pos) = divide(cell_data[i,0:2],displacement)
				cell_data[i,:] = np.array([new_pos[0], new_pos[1], 1.0, norm_sample(mu,sigma), t_elapsed, 0.0, 0.0], dtype=np.float32)
					
				#Add the child
				cell_data = np.append(cell_data, np.float32([[child_pos[0],child_pos[1],1.0,norm_sample(mu,sigma),t_elapsed,0.0,0.0]]),axis=0)
	
		#Reset Velocities to 0
		cell_data[:,5:7] = np.array([0.0,0.0], dtype=np.float32)
		
		#Draw/Output Files (Doing it at an analogous time as the parallel version)
		if args.draw:
			draw_once(cell_data[:,:3],plt, fig)
		if args.outputf != None and t_elapsed % args.freq < dt:
			file_name = args.outputf + '_' + str(t_elapsed)
			np.savetxt(file_name, cell_data,fmt='%f')
			
		# Update the forces
		for (k1, k2) in itertools.combinations(range(num_cells),2):
			c1r = cell_data[k1][0:3]
			c2r = cell_data[k2][0:3]
			c1 = c1r[:2]
			r1 = c1r[2]
			c2 = c2r[:2]
			r2 = c2r[2]
			D =  distance(c1, c2)
			if (r1 + r2 ) > D:
				#apply force here
				v_net = v_vector(r1,r2,c1,c2, D,v_const)
				cell_data[k1, 5:7] += v_net
				cell_data[k2, 5:7] -= v_net
		
		#Update position based on forces
		for i in xrange(num_cells):
			cell_data[i,0:2] += dt * cell_data[i,5:7]
			
		t_elapsed += dt
		
	return num_cells

#RUNMODE 1: BFS Serial Implementation
def simulate_bfs_serial(init_cell_data,args):
	dt = args.dt
	t_elapsed = 0
	t_end = args.time
	
	displacement = args.disp
	v_const = 25.0
	
	grow_const = args.grow
	mu = args.rmax
	sigma = args.rsd
 
	# plotting stuff
	if args.draw:
		plt.ion()
		fig = plt.figure(figsize=(16, 12)).add_subplot(111, aspect='equal')
		plt.xlim(-15,15)
		plt.ylim(-15,15)
	
	graph,cell_data,force_vectors = generate_graph(init_cell_data)
	num_cells = init_cell_data.shape[0];


	while t_elapsed < t_end:
		glob_r_max = 0
		
		#Growth and Division
		for cell, params in cell_data.items():
			(r, r_max, t_start) = params[2:]
			r = grow_const * exp((t_elapsed - t_start)/2.0)
			
			if r_max > glob_r_max:
				glob_r_max = r_max
			
			if r < r_max:
				cell_data[cell][2] = r #Grow
			else: #Divide
				num_cells += 1
				new_node = str(num_cells)
				cen = params[:2]				
				(new_pos,child_pos) = divide(cen,displacement)
				cell_data[cell] = np.array([new_pos[0], new_pos[1], 1.0, norm_sample(mu,sigma), t_elapsed])
				cell_data[new_node] =  np.array([child_pos[0], child_pos[1], 1.0, norm_sample(mu,sigma), t_elapsed])
			
				graph[cell].append(new_node)
				graph[new_node] = [cell]
				force_vectors[new_node] = np.array([0.0,0.0])
				
		#Draw/Output Files (Doing it at an analogous time as the parallel version)
		if args.draw:
			data = np.array(cell_data.values())
			draw_once(data[:,:3], plt, fig)
		if args.outputf != None and t_elapsed % args.freq < dt:
			file_name = args.outputf + '_' + str(t_elapsed)
			try: np.savetxt(file_name, cell_data,fmt='%f')
			except:
				data = np.array(cell_data.values())
				np.savetxt(file_name, data,fmt='%f')

		# Update the graph and forces
		# Reset Forces
		for node in force_vectors:
			# graph[node] = []
			force_vectors[node][0] = 0.0
			force_vectors[node][1] = 0.0

		glob_thresh = (max(0, glob_r_max - 1) + 0.11)

		graph = make_graph_bfs2(graph, cell_data, glob_thresh)

		for k1, adj_nodes in graph.iteritems():
			c1r = cell_data[k1][0:3]
			c1 = c1r[:2]
			r1 = c1r[2]
			for k2 in adj_nodes:
				c2r = cell_data[k2][0:3]
				c2 = c2r[:2]
				r2 = c2r[2]
				D =  distance(c1, c2)
				if (r1 + r2 ) > D:
					#apply force here
					v_net = v_vector(r1,r2,c1,c2,D,v_const)
					force_vectors[k1] += v_net

		#Apply velocities to update position
		for cell in cell_data:
			cell_data[cell][:2] += dt * force_vectors[cell]
	
		t_elapsed += dt
	return num_cells

#RUNMODE 2: Naive Parallel Implementation
def simulate_parallel(init_cell_data, args, comm):
	rank = comm.Get_rank()
	size = comm.Get_size()
	
	dt = args.dt
	t_elapsed = 0
	t_end = args.time
	
	displacement = args.disp
	v_const = 25.0
	
	grow_const = args.grow
	mu = args.rmax
	sigma = args.rsd
 
	# plotting stuff
	if args.draw and rank == 0:
		plt.ion()
		fig = plt.figure(figsize=(16, 12)).add_subplot(111, aspect='equal')
		plt.xlim(-15,15)
		plt.ylim(-15,15)
	
	cell_data = init_cell_data
	num_feats = cell_data.shape[1];
		
	while t_elapsed < t_end:
		num_cells = cell_data.shape[0];
		#if rank == 0:
		#	print t_elapsed, t_elapsed/dt, num_cells 
		
		N = int(np.ceil(float(num_cells) / size)) #Number of cells per process
		
		start_index = rank * N
		end_index = min(start_index+N, num_cells)
		
		
		#Local copy of the cells
		local_cell_data = cell_data[start_index:end_index,:]
								
		for i in xrange(local_cell_data.shape[0]):
			#Grow and Divide
			r = grow_const * exp((t_elapsed - local_cell_data[i,4])/2.0) #Update the radius						

			if r < local_cell_data[i,3]:
				local_cell_data[i,2] = r #Grow
			else:
				#Divide
				(new_pos,child_pos) = divide(local_cell_data[i,0:2],displacement)
				local_cell_data[i,0:2] = new_pos
				
				local_cell_data[i,2] = 1 #New Radius
				local_cell_data[i,3] = norm_sample(mu,sigma) #New division time
				local_cell_data[i,4] = t_elapsed #New Start Time
			
				#Add the child
				local_cell_data = np.append(local_cell_data,np.float32([[child_pos[0],child_pos[1],1.0,norm_sample(mu,sigma),t_elapsed,0.0,0.0]]),axis=0)
			
		#Compile all of the data
		local_num_cells = np.array([local_cell_data.shape[0]],dtype=np.int32)
		all_num_cells = np.empty(size,dtype=np.int32)
		
		comm.Allgather(local_num_cells,all_num_cells)
		
		scanned_indices = np.append(np.array([0],dtype=np.int32),np.cumsum(all_num_cells),axis=0)
		num_cells = scanned_indices[-1]
		
		cell_data = np.empty((num_cells,num_feats),dtype=np.float32)
		comm.Allgatherv(local_cell_data,[cell_data,tuple(num_feats*all_num_cells),tuple(num_feats*scanned_indices[0:-1]),MPI.FLOAT])
		
		#Draw/Output Files
		if args.draw and rank == 0:
			draw_once(cell_data[:,:3],plt, fig)
		if args.outputf != None and t_elapsed % args.freq < dt and rank == 0:
			file_name = args.outputf + '_' + str(t_elapsed)
			np.savetxt(file_name, cell_data,fmt='%f')
	
		#Reset velocities
		cell_data[start_index:end_index,5:7] = np.array([0.0,0.0])
		
		N = int(np.ceil(float(num_cells) / size)) #Number of cells per process
		start_index = rank * N
		end_index = min(start_index+N, num_cells)
		
		#Compute forces
		for i in xrange(start_index,end_index):
			for j in xrange(i,num_cells):
				if j > i or j < start_index:
					c1r = cell_data[i][0:3]
					c2r = cell_data[j][0:3]
					c1 = c1r[:2]
					r1 = c1r[2]
					c2 = c2r[:2]
					r2 = c2r[2]
					D =  distance(c1, c2)
					
					if D < r1+r2:
						v_net = v_vector(r1,r2,c1,c2, D, v_const)
						cell_data[i,5:7] += v_net
						cell_data[j,5:7] -= v_net
			
			#Update position based on velocity
			cell_data[i,0:2] += dt * cell_data[i,5:7]
		
		t_elapsed += dt
		
	return num_cells

#MAIN METHOD	
if __name__ == '__main__':
	parser = argparse.ArgumentParser(description='Simulate circular cells growing out on a petri dish.')
	
	parser.add_argument('time',type=float,help='How long to run simulation for')
	
	parser.add_argument('-i','--inputf',help='Provides an input file for any initial conditions. Should be whitespace delimited and be of the format [X Y Radius]. Otherwise defaults to single cell centered at the origin with radius 1.',default=None)
	parser.add_argument('-o','--outputf',help='Output File Name. Outputs to a text file <outputf>_<time>. Default doesn\'t output')
	parser.add_argument('-f','--freq',help='Frequency of file outputs. Defaults to 0.5',type=float,default=0.5)
	
	parser.add_argument('-d','--draw',dest='draw',action='store_true',help='Draws a representation of the colony with each iteration.')
	parser.set_defaults(draw=False)
		
	parser.add_argument('-r','--runmode',type=int, default=0, choices=[0, 1,2], help='0: Serial, 1: Serial with BFS, 2: Naive Parallel Implementation. Defaults to 0')
	parser.add_argument('-g','--grow',type=float,default=1.1,help='Growth Rate. Defaults to 1.1')
	parser.add_argument('--rmax',type=float,default=1.5,help='Mean maximum radius before division occurs. Defaults to 1.5')
	parser.add_argument('--rsd',type=float,default=0.1,help='Standard Deviation of the maximum radius before division occurs. Defaults to 0.1')
	parser.add_argument('--disp',type=float,default=0.1,help='Displacement between cells when dividing. Defaults to 0.1')
	
	parser.add_argument('--dt',type=float,default=0.01,help='Time between iterations')
	
	
	args = parser.parse_args()
	
	comm = MPI.COMM_WORLD
	rank = comm.Get_rank()
	size = comm.Get_size()

	#Initial Conditions
	if args.inputf == None:
		#cell_data = (C_x,C_y,r,r_max,t_start,v_x,v_y)
		cell_data = np.array([[0.0, 0.0, 1.0, norm_sample(args.rmax,args.rsd), 0.0,0.0,0.0]],dtype=np.float32)
	else:
		start_pos_r = np.loadtxt(args.inputf,dtype=np.float32,ndmin=2) #Read initial position and radius from the file
		num_cells = start_pos_r.shape[0]
		
		start_tdiv = np.random.normal(loc=args.rmax, scale = args.rsd, size = (num_cells,1)) #Division Radius
		
		start_tstart_v = np.zeros((num_cells,3),dtype=np.float32) #Starting time, and initial velocities
		
		#Concatenate the final result
		cell_data = np.float32(np.concatenate((start_pos_r,start_tdiv,start_tstart_v),axis=1))
		
	
	
	if args.runmode == 0: #Naive Serial
		start_time = time.time()
		num_cells = simulate_naive_serial(cell_data,args)
		end_time = time.time()
		print "Naive Serial Processing Time ending with %i cells: %f" % (num_cells, (end_time - start_time))
	elif args.runmode == 1: #BFS Serial
		start_time = time.time()
		num_cells = simulate_bfs_serial(cell_data,args)
		end_time = time.time()
		print "Serial BFS Processing Time ending with %i cells: %f" % (num_cells, (end_time - start_time))
	elif args.runmode == 2: #Naive Parallel
		comm.Bcast(cell_data,root=0)
		comm.barrier()
		start_time = MPI.Wtime()
		num_cells = simulate_parallel(cell_data, args, comm)
		end_time = MPI.Wtime()
		if rank == 0:
			print "Parallel Processing Time on %i processes ending with %i cells: %f" % (size,num_cells, (end_time - start_time))

